/*
 * vim:sw=2 ts=2 sts=2:et ai:bg=dark
  mpbz2 - a program to do multi-threaded bzip2 compression using libbzip2
  
  Copyright Clint Byrum, 2004. All Rights Reserved.
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ---

  $Id: mpbz2.c 42 2004-08-27 23:55:33Z bronaugh $
  
 */

#include <getopt.h>

#include "mpbz2.h"
#include "thread_funcs.h"
#include "helpers.h"
#include "ring_buffer.h"

int main(int argc,char *argv[]) {
  int i;
  int val;
  int compress_level = DEFAULT_COMPRESS_LEVEL;
  int num_compress_threads = DEFAULT_COMPRESS_THREADS;
  int buf_size = DEFAULT_BUF_SIZE;
  mpbz2_info* inf;

  pthread_t* compress_threads;
  pthread_t writer_thread;
  pthread_t reader_thread;
  char* optstring = "s:o:j:c";

  inf = (mpbz2_info*)malloc(sizeof(mpbz2_info));

  inf->instream = stdin;
  inf->outstream = stdout;

  /* Parse args */
  if(argc > 1) {
    opterr = 0;
    while ((val = getopt(argc, argv, optstring)) > 0) {
      switch (val) {
      case 'j':
	num_compress_threads = atoi(optarg);
	break;
      case 's':
	buf_size = atoi(optarg);
	break;
      case 'c':
	inf->outstream = stdout;
	break;
      case 'o':
	/* Open file for output */
	if (!(inf->outstream = fopen(optarg, "w"))) {
	  fprintf(stderr, "Couldn't open output file \"%s\"!\n", optarg);
	  exit(1);
	}
	break;
      case '?':
	if(isdigit(optopt) && optopt > (int)'0') {
	  compress_level = 10 - (optopt - (int)'0');
	}
	break;
	
      }
    }

    if (argc - optind < 0) {
      logMsg(LOG_ERROR, "usage: mpbz2 [-s bufsize] [-j nproc] [-c] [-o file ] [-1-9] <input-file>\n");
      exit(1);
    }
    
    
    /* For now input can only be instream */
    if (!strcmp(argv[optind], "-")) {
      inf->instream=stdin;
    } else {
      if (!(inf->instream = fopen(argv[optind], "r"))) {
	fprintf(stderr, "Couldn't open input file \"%s\"!\n", argv[optind]);
	exit(1);
      }
    }
  }

  compress_threads = (pthread_t*) malloc(sizeof(pthread_t) * num_compress_threads);

  inf->buf = mpbz2_create_ring_buffer(buf_size, compress_level * BLOCK_SIZE);
  inf->read_threads = inf->write_threads = inf->compress_threads = 0;
  inf->started_read = inf->started_write = inf->started_compress = false;
  inf->done = 0;

  pthread_mutex_init(&inf->lock, NULL);

  pthread_cond_init(&inf->cond_compress, NULL);
  pthread_cond_init(&inf->cond_read, NULL);
  pthread_cond_init(&inf->cond_write, NULL);

  pthread_create(&reader_thread, NULL, readController, inf);
  for (i = 0; i < num_compress_threads; i++) {
    pthread_create(&compress_threads[i], NULL, compressController, inf);
  }
  pthread_create(&writer_thread, NULL, writeController, inf);

  pthread_join(reader_thread,NULL);

  for(i = 0; i < num_compress_threads; i++) {
    pthread_join(compress_threads[i], NULL);
  }
  logMsg(LOG_INFO,"compress threads exited\n");
  
  pthread_join(writer_thread,NULL);

  pthread_mutex_destroy(&inf->lock);
  pthread_cond_destroy(&inf->cond_compress);
  pthread_cond_destroy(&inf->cond_read);
  pthread_cond_destroy(&inf->cond_write);

  mpbz2_destroy_ring_buffer(inf->buf);
  logMsg(LOG_INFO,"main thread exiting\n");
  return 0;
}
